/* Talika panel applet: Switch between open windows using icons.
 * (C) 2009-2010 Sinew Software Systems
	 *
 * Authors:  Vinod Kumar Dahiya
 *           Hemant Kumar
 *
 * NOTE: Code has been taken from libwnck and metacity sources.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
	 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "xutils.h"
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "inlinepixbufs.h"
#include <X11/extensions/Xcomposite.h>

/*######################### Get icon from X11 ################################*/

void
_wnck_error_trap_push (void)
{
	gdk_error_trap_push ();
}

int
_wnck_error_trap_pop (void)
{
	XSync (gdk_display, False);
	return gdk_error_trap_pop ();
}

/* The icon-reading code is copied
 * from metacity, please sync bugfixes
 */
static gboolean
find_largest_sizes (gulong * data, gulong nitems, int *width, int *height)
{
	*width = 0;
	*height = 0;

	while (nitems > 0)
	{
		int w, h;
		gboolean replace;

		replace = FALSE;

		if (nitems < 3)
			return FALSE;	/* no space for w, h */

		w = data[0];
		h = data[1];

		if (nitems < ((w * h) + 2))
			return FALSE;	/* not enough data */

		*width = MAX (w, *width);
		*height = MAX (h, *height);

		data += (w * h) + 2;
		nitems -= (w * h) + 2;
	}

	return TRUE;
}

static gboolean
find_best_size (gulong * data,
                gulong nitems,
                int ideal_width,
                int ideal_height, int *width, int *height, gulong ** start)
{
	int best_w;
	int best_h;
	gulong *best_start;
	int max_width, max_height;

	*width = 0;
	*height = 0;
	*start = NULL;

	if (!find_largest_sizes (data, nitems, &max_width, &max_height))
		return FALSE;

	if (ideal_width < 0)
		ideal_width = max_width;
	if (ideal_height < 0)
		ideal_height = max_height;

	best_w = 0;
	best_h = 0;
	best_start = NULL;

	while (nitems > 0)
	{
		int w, h;
		gboolean replace;

		replace = FALSE;

		if (nitems < 3)
			return FALSE;	/* no space for w, h */

		w = data[0];
		h = data[1];

		if (nitems < ((w * h) + 2))
			break;	/* not enough data */

		if (best_start == NULL)
		{
			replace = TRUE;
		}
		else
		{
			/* work with averages */
			const int ideal_size =
				(ideal_width + ideal_height) / 2;
			int best_size = (best_w + best_h) / 2;
			int this_size = (w + h) / 2;

			/* larger than desired is always better than smaller */
			if (best_size < ideal_size && this_size >= ideal_size)
				replace = TRUE;
			/* if we have too small, pick anything bigger */
			else if (best_size < ideal_size &&
			         this_size > best_size)
				replace = TRUE;
			/* if we have too large, pick anything smaller
			 * but still >= the ideal
			 */
			else if (best_size > ideal_size &&
			         this_size >= ideal_size &&
			         this_size < best_size)
				replace = TRUE;
		}

		if (replace)
		{
			best_start = data + 2;
			best_w = w;
			best_h = h;
		}

		data += (w * h) + 2;
		nitems -= (w * h) + 2;
	}

	if (best_start)
	{
		*start = best_start;
		*width = best_w;
		*height = best_h;
		return TRUE;
	}
	else
		return FALSE;
}

static void
argbdata_to_pixdata (gulong * argb_data, int len, guchar ** pixdata)
{
	guchar *p;
	int i;

	*pixdata = g_new (guchar, len * 4);
	p = *pixdata;

	/* One could speed this up a lot. */
	i = 0;
	while (i < len)
	{
		guint argb;
		guint rgba;

		argb = argb_data[i];
		rgba = (argb << 8) | (argb >> 24);

		*p = rgba >> 24;
		++p;
		*p = (rgba >> 16) & 0xff;
		++p;
		*p = (rgba >> 8) & 0xff;
		++p;
		*p = rgba & 0xff;
		++p;

		++i;
	}
}

static gboolean
read_rgb_icon (Window xwindow,
               int ideal_width,
               int ideal_height,
               int ideal_mini_width,
               int ideal_mini_height,
               int *width,
               int *height,
               guchar ** pixdata,
               int *mini_width, int *mini_height, guchar ** mini_pixdata)
{
	Atom type;
	int format;
	gulong nitems;
	gulong bytes_after;
	int result, err;
	gulong *data;
	gulong *best;
	int w, h;
	gulong *best_mini;
	int mini_w, mini_h;

	_wnck_error_trap_push ();
	type = None;
	data = NULL;
	result = XGetWindowProperty (gdk_display,
	                             xwindow,
	                             _wnck_atom_get ("_NET_WM_ICON"),
	                             0, G_MAXLONG,
	                             False, XA_CARDINAL, &type, &format,
	                             &nitems, &bytes_after, (void *) &data);

	err = _wnck_error_trap_pop ();

	if (err != Success || result != Success)
		return FALSE;

	if (type != XA_CARDINAL)
	{
		XFree (data);
		return FALSE;
	}

	if (!find_best_size (data, nitems,
	                     ideal_width, ideal_height, &w, &h, &best))
	{
		XFree (data);
		return FALSE;
	}

	if (!find_best_size (data, nitems,
	                     ideal_mini_width, ideal_mini_height,
	                     &mini_w, &mini_h, &best_mini))
	{
		XFree (data);
		return FALSE;
	}

	*width = w;
	*height = h;

	*mini_width = mini_w;
	*mini_height = mini_h;

	argbdata_to_pixdata (best, w * h, pixdata);
	argbdata_to_pixdata (best_mini, mini_w * mini_h, mini_pixdata);

	XFree (data);

	return TRUE;
}

static void
free_pixels (guchar * pixels, gpointer data)
{
	g_free (pixels);
}

static GdkPixbuf *
scaled_from_pixdata (guchar * pixdata, int w, int h, int new_w, int new_h)
{
	GdkPixbuf *src;
	GdkPixbuf *dest;

	src = gdk_pixbuf_new_from_data (pixdata,
	                                GDK_COLORSPACE_RGB,
	                                TRUE,
	                                8, w, h, w * 4, free_pixels, NULL);

	if (src == NULL)
		return NULL;

	if (w != h)
	{
		GdkPixbuf *tmp;
		int size;

		size = MAX (w, h);

		tmp = gdk_pixbuf_new (GDK_COLORSPACE_RGB, TRUE, 8, size,
		                      size);

		if (tmp != NULL)
		{
			gdk_pixbuf_fill (tmp, 0);
			gdk_pixbuf_copy_area (src, 0, 0, w, h,
			                      tmp,
			                      (size - w) / 2, (size - h) / 2);

			g_object_unref (src);
			src = tmp;
		}
	}

	if (w != new_w || h != new_h)
	{
		dest = gdk_pixbuf_scale_simple (src, new_w, new_h,
		                                GDK_INTERP_HYPER);

		g_object_unref (G_OBJECT (src));
	}
	else
	{
		dest = src;
	}

	return dest;
}

//get icon from x server, null if failed
GdkPixbuf *
talika_get_icon_from_xserver (WnckWindow * window, guint size)
{
	GdkPixbuf *icon;
	GdkPixbuf *mini_icon;
	guchar *pixdata;
	int w, h;
	guchar *mini_pixdata;
	int mini_w, mini_h;

	gulong xid = wnck_window_get_xid (window);

	icon = NULL;
	mini_icon = NULL;

	if (read_rgb_icon (xid,
	                   size, size,
	                   16, 16,
	                   &w, &h, &pixdata, &mini_w, &mini_h, &mini_pixdata))
	{
		icon = scaled_from_pixdata (pixdata, w, h, size,
		                            size);

		mini_icon = scaled_from_pixdata (mini_pixdata, mini_w, mini_h,
		                                 16,
		                                 16);
		g_object_unref (mini_icon);
		return icon;
	}

	return NULL;
}


/*######################### Get window picture ###############################*/

static GdkPixbuf *
meta_ui_get_pixbuf_from_pixmap (Pixmap   pmap)
{
	GdkPixmap *gpmap;
	GdkScreen *screen;
	GdkPixbuf *pixbuf;
	GdkColormap *cmap;
	int width, height, depth;

	gpmap = gdk_pixmap_foreign_new (pmap);
	screen = gdk_drawable_get_screen (gpmap);

	gdk_drawable_get_size (GDK_DRAWABLE (gpmap), &width, &height);

	depth = gdk_drawable_get_depth (GDK_DRAWABLE (gpmap));
	if (depth <= 24)
		cmap = gdk_screen_get_rgb_colormap (screen);
	else
		cmap = gdk_screen_get_rgba_colormap (screen);

	pixbuf = gdk_pixbuf_get_from_drawable (NULL, gpmap, cmap, 0, 0, 0, 0,
	                                       width, height);

	g_object_unref (gpmap);

	return pixbuf;
}

static GdkPixbuf *
get_window_pixbuf (Display *dpy, Window window, gint size, int *width, int *height)
{
	Pixmap pmap;
	GdkPixbuf *pixbuf, *scaled;
	double ratio;
	double MAX_PREVIEW_SIZE = size;

	gdk_error_trap_push ();
	/* ... Call the X function which may cause an error here ... */
	pmap = XCompositeNameWindowPixmap (dpy, window);
	/* Flush the X queue to catch errors now. */
	gdk_flush ();
	if (gdk_error_trap_pop ())
	{
		/* ... Handle the error here ... */
		g_print ("X error getting preview\n");
		pmap = None;
	}

	if (pmap == None||pmap == BadMatch)
		return NULL;

	pixbuf = meta_ui_get_pixbuf_from_pixmap (pmap);
	XFreePixmap(dpy,pmap);

	if (pixbuf == NULL) 
		return NULL;

	*width = gdk_pixbuf_get_width (pixbuf);
	*height = gdk_pixbuf_get_height (pixbuf);

	/* Scale pixbuf to max dimension MAX_PREVIEW_SIZE */
	/*if (*width > *height)
	{
		ratio = ((double) *width) / MAX_PREVIEW_SIZE;
		*width = (int) MAX_PREVIEW_SIZE;
		*height = (int) (((double) *height) / ratio);
	}
	else*/
	{
		ratio = ((double) *height) / MAX_PREVIEW_SIZE;
		*height = (int) MAX_PREVIEW_SIZE;
		*width = (int) (((double) *width) / ratio);
	}
	
	if(*width > *height*2)
	{
		ratio = ((double) *width) / MAX_PREVIEW_SIZE;
		*width = (int) MAX_PREVIEW_SIZE*1.5;
		*height = (int) (((double) *height) / ratio);
	}

	scaled = gdk_pixbuf_scale_simple (pixbuf, *width, *height,
	                                  GDK_INTERP_BILINEAR);
	g_object_unref (pixbuf);
	return scaled;
}

GdkPixbuf *
talika_get_window_picture (WnckWindow * window, guint *size)
{
	Display *display = gdk_x11_display_get_xdisplay (
	                                                 gdk_display_get_default ());
	Window win = wnck_window_get_xid (window);

	gint h=0,w=0;

	GdkPixbuf *buf = get_window_pixbuf (display,win,*size,&w,&h);
	*size = w;

	return buf;
}


/*######################### Utility functions ################################*/

//make a GtkLabel bold
void 
_make_gtk_label_bold (GtkLabel * label)
{
	PangoFontDescription *font_desc;

	font_desc = pango_font_description_new ();

	pango_font_description_set_weight (font_desc, PANGO_WEIGHT_BOLD);

	/* This will only affect the weight of the font, the rest is
	 * from the current state of the widget, which comes from the
	 * theme or user prefs, since the font desc only has the
	 * weight flag turned on.
	 */
	gtk_widget_modify_font (GTK_WIDGET (label), font_desc);

	pango_font_description_free (font_desc);
}

//get Gdk screen corsponding to Wnck screen
GdkScreen*
talika_screen_get_gdk_screen (WnckScreen * screen)
{
	g_return_val_if_fail (WNCK_IS_SCREEN (screen), NULL);

	return gdk_display_get_screen (gdk_display_get_default (),
	                               wnck_screen_get_number (screen));
}

/* stock icon code Copyright (C) 2002 Jorn Baayen <jorn@nl.linux.org> */
typedef struct
{
  char *stock_id;
  const guint8 *icon_data;
} StockIcon;

void
talika_stock_icons_init (void)
{
  GtkIconFactory *factory;
  int i;
  static gboolean done = FALSE;

  StockIcon items[] =
  {
    { TALIKA_STOCK_CLOSE,   stock_delete_data   },
    { TALIKA_STOCK_MINIMIZE, stock_minimize_data },
    { TALIKA_STOCK_MAXIMIZE, stock_maximize_data },
	{ TALIKA_STOCK_LAUNCH,   stock_launch_data   },
    { TALIKA_STOCK_PIN, stock_pin_data },
    { TALIKA_STOCK_UNPIN, stock_unpin_data },
	{ TALIKA_STOCK_CHANGE_ICON, stock_change_icon_data }
  };

  if (done)
    return;

  done = TRUE;
  
  factory = gtk_icon_factory_new ();
  gtk_icon_factory_add_default (factory);

  for (i = 0; i < (gint) G_N_ELEMENTS (items); i++)
    {
      GtkIconSet *icon_set;
      GdkPixbuf *pixbuf;

      pixbuf = gdk_pixbuf_new_from_inline (-1, items[i].icon_data,
					   FALSE,
					   NULL);

      icon_set = gtk_icon_set_new_from_pixbuf (pixbuf);
      gtk_icon_factory_add (factory, items[i].stock_id, icon_set);
      gtk_icon_set_unref (icon_set);
		
      g_object_unref (G_OBJECT (pixbuf));
    }

  g_object_unref (G_OBJECT (factory));
}

/*########################### Zoom Animation #################################*/
/* Zoom animation */
#define MINIATURIZE_ANIMATION_FRAMES_Z   1
#define MINIATURIZE_ANIMATION_STEPS_Z    6
/* the delay per draw */
#define MINIATURIZE_ANIMATION_DELAY_Z    10

/* zoom factor, steps and delay if composited (factor must be odd) */
#define ZOOM_FACTOR 5
#define ZOOM_STEPS  14
#define ZOOM_DELAY 10

typedef struct {
	int size;
	int size_start;
	int size_end;
	PanelAppletOrient orientation;
	double opacity;
	GdkPixbuf *pixbuf;
	guint timeout_id;
} CompositedZoomData;

static gboolean
zoom_timeout (GtkWidget *window)
{
	gtk_widget_queue_draw (window);
	return TRUE;
}

static gboolean
zoom_expose (GtkWidget      *widget,
	     GdkEventExpose *event,
	     gpointer        user_data)
{
	CompositedZoomData *zoom;

	zoom = user_data;

	if (zoom->size >= zoom->size_end) {
		if (zoom->timeout_id)
			g_source_remove (zoom->timeout_id);
		zoom->timeout_id = 0;

		g_object_unref (zoom->pixbuf);
		zoom->pixbuf = NULL;

		g_slice_free (CompositedZoomData, zoom);

		gtk_widget_destroy (widget);
	} else {
		GdkPixbuf *scaled;
		int width, height;
		int x = 0, y = 0;
		cairo_t *cr;

		gtk_window_get_size (GTK_WINDOW (widget), &width, &height);

		zoom->size += MAX ((zoom->size_end - zoom->size_start) / ZOOM_STEPS, 1);
		zoom->opacity -= 1.0 / ((double) ZOOM_STEPS + 1);

		scaled = gdk_pixbuf_scale_simple (zoom->pixbuf,
						  zoom->size, zoom->size,
						  GDK_INTERP_BILINEAR);

		switch (zoom->orientation) {
		case PANEL_APPLET_ORIENT_DOWN:
			x = (width - gdk_pixbuf_get_width (scaled)) / 2;
			y = 0;
			break;

		case PANEL_APPLET_ORIENT_LEFT:
			x = width - gdk_pixbuf_get_width (scaled);
			y = (height - gdk_pixbuf_get_height (scaled)) / 2;
			break;

		case PANEL_APPLET_ORIENT_UP:
			x = (width - gdk_pixbuf_get_width (scaled)) / 2;
			y = height - gdk_pixbuf_get_height (scaled);
			break;

		case PANEL_APPLET_ORIENT_RIGHT:
			x = 0;
			y = (height - gdk_pixbuf_get_height (scaled)) / 2;
			break;
		}


		cr = gdk_cairo_create (widget->window);
		cairo_set_operator (cr, CAIRO_OPERATOR_SOURCE);
		cairo_set_source_rgba (cr, 0, 0, 0, 0.0);
		cairo_rectangle (cr, 0, 0, width, height);
		cairo_fill (cr);

		gdk_cairo_set_source_pixbuf (cr, scaled, x, y);
		cairo_set_operator (cr, CAIRO_OPERATOR_OVER);
		cairo_paint_with_alpha (cr, MAX (zoom->opacity, 0));

		cairo_destroy (cr);
		g_object_unref (scaled);
	}

	return FALSE;
}

static void 
draw_zoom_animation_composited (GdkScreen *gscreen,
				int x, int y, int w, int h,
				GdkPixbuf *pixbuf,
				PanelAppletOrient orientation)
{
	GtkWidget *win;
	CompositedZoomData *zoom;
	int wx = 0, wy = 0;

	w += 2;
	h += 2;

	zoom = g_slice_new (CompositedZoomData);
	zoom->size = w;
	zoom->size_start = w;
	zoom->size_end = w * ZOOM_FACTOR;
	zoom->orientation = orientation;
	zoom->opacity = 1.0;
	zoom->pixbuf = g_object_ref (pixbuf);
	zoom->timeout_id = 0;

	win = gtk_window_new (GTK_WINDOW_POPUP);

	gtk_window_set_screen (GTK_WINDOW (win), gscreen);
	gtk_window_set_keep_above (GTK_WINDOW (win), TRUE);
	gtk_window_set_decorated (GTK_WINDOW (win), FALSE);
	gtk_widget_set_app_paintable(win, TRUE);
	gtk_widget_set_colormap (win, gdk_screen_get_rgba_colormap (gscreen));

	gtk_window_set_gravity (GTK_WINDOW (win), GDK_GRAVITY_STATIC);
	gtk_window_set_default_size (GTK_WINDOW (win),
				     w * ZOOM_FACTOR, h * ZOOM_FACTOR);

	switch (zoom->orientation) {
	case PANEL_APPLET_ORIENT_DOWN:
		wx = x - w * (ZOOM_FACTOR / 2);
		wy = y;
		break;

	case PANEL_APPLET_ORIENT_LEFT:
		wx = x - w * (ZOOM_FACTOR - 1);
		wy = y - h * (ZOOM_FACTOR / 2);
		break;

	case PANEL_APPLET_ORIENT_UP:
		wx = x - w * (ZOOM_FACTOR / 2);
		wy = y - h * (ZOOM_FACTOR - 1);
		break;

	case PANEL_APPLET_ORIENT_RIGHT:
		wx = x;
		wy = y - h * (ZOOM_FACTOR / 2);
		break;
	}

	gtk_window_move (GTK_WINDOW (win), wx, wy);

	g_signal_connect (G_OBJECT (win), "expose-event",
			  G_CALLBACK (zoom_expose), zoom);

	/* see doc for gtk_widget_set_app_paintable() */
	gtk_widget_realize (win);
	gdk_window_set_back_pixmap (win->window, NULL, FALSE);
	gtk_widget_show (win);

	zoom->timeout_id = g_timeout_add (ZOOM_DELAY,
					  (GSourceFunc) zoom_timeout,
					  win);
}

static void 
draw_zoom_animation (GdkScreen *gscreen,
		     int x, int y, int w, int h,
		     int fx, int fy, int fw, int fh, 
		     int steps)
{
#define FRAMES (MINIATURIZE_ANIMATION_FRAMES_Z)
	float cx[FRAMES], cy[FRAMES], cw[FRAMES], ch[FRAMES];
	int cxi[FRAMES], cyi[FRAMES], cwi[FRAMES], chi[FRAMES];
	float xstep, ystep, wstep, hstep;
	int i, j;
	GC frame_gc;
	XGCValues gcv;
	GdkColor color = { 65535, 65535, 65535 };
	Display *dpy;
	Window root_win;
	int screen;
	int depth;

	dpy = gdk_x11_display_get_xdisplay (gdk_screen_get_display (gscreen));
	root_win = gdk_x11_drawable_get_xid (gdk_screen_get_root_window (gscreen));
	screen = gdk_screen_get_number (gscreen);
	depth = gdk_drawable_get_depth (gdk_screen_get_root_window (gscreen));

	/* frame GC */
	gdk_colormap_alloc_color (
		gdk_screen_get_system_colormap (gscreen), &color, FALSE, TRUE);
	gcv.function = GXxor;
	/* this will raise the probability of the XORed color being different
	 * of the original color in PseudoColor when not all color cells are
	 * initialized */
	if (DefaultVisual(dpy, screen)->class==PseudoColor)
		gcv.plane_mask = (1<<(depth-1))|1;
	else
		gcv.plane_mask = AllPlanes;
	gcv.foreground = color.pixel;
	if (gcv.foreground == 0)
		gcv.foreground = 1;
	gcv.line_width = 1;
	gcv.subwindow_mode = IncludeInferiors;
	gcv.graphics_exposures = False;

	frame_gc = XCreateGC(dpy, root_win, GCForeground|GCGraphicsExposures
			     |GCFunction|GCSubwindowMode|GCLineWidth
			     |GCPlaneMask, &gcv);

	xstep = (float)(fx-x)/steps;
	ystep = (float)(fy-y)/steps;
	wstep = (float)(fw-w)/steps;
	hstep = (float)(fh-h)/steps;

	for (j=0; j<FRAMES; j++) {
		cx[j] = (float)x;
		cy[j] = (float)y;
		cw[j] = (float)w;
		ch[j] = (float)h;
		cxi[j] = (int)cx[j];
		cyi[j] = (int)cy[j];
		cwi[j] = (int)cw[j];
		chi[j] = (int)ch[j];
	}
	XGrabServer(dpy);
	for (i=0; i<steps; i++) {
		for (j=0; j<FRAMES; j++) {
			XDrawRectangle(dpy, root_win, frame_gc, cxi[j], cyi[j], cwi[j], chi[j]);
		}
		XFlush(dpy);
#if (MINIATURIZE_ANIMATION_DELAY_Z > 0)
		usleep(MINIATURIZE_ANIMATION_DELAY_Z);
#else
		usleep(10);
#endif
		for (j=0; j<FRAMES; j++) {
			XDrawRectangle(dpy, root_win, frame_gc, 
				       cxi[j], cyi[j], cwi[j], chi[j]);
			if (j<FRAMES-1) {
				cx[j]=cx[j+1];
				cy[j]=cy[j+1];
				cw[j]=cw[j+1];
				ch[j]=ch[j+1];

				cxi[j]=cxi[j+1];
				cyi[j]=cyi[j+1];
				cwi[j]=cwi[j+1];
				chi[j]=chi[j+1];

			} else {
				cx[j]+=xstep;
				cy[j]+=ystep;
				cw[j]+=wstep;
				ch[j]+=hstep;

				cxi[j] = (int)cx[j];
				cyi[j] = (int)cy[j];
				cwi[j] = (int)cw[j];
				chi[j] = (int)ch[j];
			}
		}
	}

	for (j=0; j<FRAMES; j++) {
		XDrawRectangle(dpy, root_win, frame_gc, 
				       cxi[j], cyi[j], cwi[j], chi[j]);
	}
	XFlush(dpy);
#if (MINIATURIZE_ANIMATION_DELAY_Z > 0)
	usleep(MINIATURIZE_ANIMATION_DELAY_Z);
#else
	usleep(10);
#endif
	for (j=0; j<FRAMES; j++) {
		XDrawRectangle(dpy, root_win, frame_gc, 
				       cxi[j], cyi[j], cwi[j], chi[j]);
	}

	XUngrabServer(dpy);
	XFreeGC (dpy, frame_gc);
	gdk_colormap_free_colors (gdk_screen_get_system_colormap (gscreen),
				  &color, 1);

}
#undef FRAMES

void
xstuff_zoom_animate (GtkWidget *widget,
		     GdkPixbuf *pixbuf,
		     PanelAppletOrient orientation,
		     GdkRectangle *opt_rect)
{
	GdkScreen *gscreen;
	GdkRectangle rect, dest;
	int monitor;

	if (opt_rect)
		rect = *opt_rect;
	else {
		gdk_window_get_origin (widget->window, &rect.x, &rect.y);
		if (GTK_WIDGET_NO_WINDOW (widget)) {
			rect.x += widget->allocation.x;
			rect.y += widget->allocation.y;
		}
		rect.height = widget->allocation.height;
		rect.width = widget->allocation.width;
	}

	gscreen = gtk_widget_get_screen (widget);

	if (gdk_screen_is_composited (gscreen) && pixbuf)
		draw_zoom_animation_composited (gscreen,
						rect.x, rect.y,
						rect.width, rect.height,
						pixbuf, orientation);
	else {
		monitor = gdk_screen_get_monitor_at_window (gscreen,
							    widget->window);
		gdk_screen_get_monitor_geometry (gscreen, monitor, &dest);

		draw_zoom_animation (gscreen,
				     rect.x, rect.y, rect.width, rect.height,
				     dest.x, dest.y, dest.width, dest.height,
				     MINIATURIZE_ANIMATION_STEPS_Z);
	}

	if (pixbuf)
		g_object_unref (pixbuf);
}

/*########################### Preview Cache ##################################*/

static void
talika_preview_image_cache_value_free(gpointer value)
{
	GdkPixbuf *buf = value;
	g_object_unref (buf);
}

TalikaPreviewImageCache* 
talika_preview_image_cache_new(void)
{
	TalikaPreviewImageCache *cache;
	cache = g_new0(TalikaPreviewImageCache,1);
	cache->hash = g_hash_table_new_full (NULL,NULL,NULL,
	                                     talika_preview_image_cache_value_free);
	return cache;
}

void
talika_preview_image_cache_free(TalikaPreviewImageCache *cache)
{
	if(!cache)
		return;
	
	g_hash_table_unref (cache->hash);
}

void
talika_preview_image_cache_update_window (TalikaPreviewImageCache *cache,
                                          WnckWindow *window,
                                          guint preview_size,
                                          GtkWidget *widget)
{
	if(!cache)
	{
		g_print ("Cache is null\n");
		return;
	}
	
	//return if screen not composited
	if( !gdk_screen_is_composited (gtk_widget_get_screen
	                               (widget)) )
	{
		g_print ("Screen not composited\n");
		return;
	}

	GdkPixbuf *buf=NULL;
	guint size=preview_size;

	if (!wnck_window_is_minimized (window))
	{
		buf = talika_get_window_picture (window, &size);
	}

	if(buf)
	{
		g_hash_table_replace (cache->hash,window,buf);
	}
}

GdkPixbuf*
talika_preview_image_cache_get_image (TalikaPreviewImageCache *cache,
                                      WnckWindow *window,
                                      guint *size)
{
	if(!cache)
	{
		g_print ("Cache is null\n");
		return NULL;
	}
	
	GdkPixbuf *temp=NULL;
	GdkPixbuf *buf=NULL;

	temp = g_hash_table_lookup (cache->hash,window);

	if(temp)
	{
		*size = gdk_pixbuf_get_width (temp);
		buf = gdk_pixbuf_copy (temp);
	}

	return buf;
}

void
talika_preview_image_cache_free_window (TalikaPreviewImageCache *cache,
                                        WnckWindow *window)
{
	if(!cache)
		return;
	
	g_hash_table_remove (cache->hash,window);
}
